/*
 * ----------------------------------------------------------------------
 * Modify, maintain and query the compatibility list
 * ----------------------------------------------------------------------
 */

#include <tcl.h>
#include <stdlib.h>
#include "tclobj.h"
#include "compat.h"
#include "register.h"
#include "signature.h"


/*
 * cvs version
 */
char * tclobj_compat_id = "$Id: compat.cc,v 1.3 1997/03/10 09:23:10 fp Exp $";


static TclObj_Comptree *
TclObj_AddComptree (Tcl_Interp *interp, TclObj_Comptree *here,
		    const TclObj_Info *compatinfo, int *result)
{
  TclObj_Comptree *node;

  if (here == NULL) {
    if ((node = new TclObj_Comptree) == NULL) {
      Tcl_AppendResult (interp, "out of memory in TclObj_AddComptree\n",
			NULL);
      *result = TCL_ERROR;
      return NULL;
    }
    node->left = node->right = NULL;
    node->info = compatinfo;
    return node;
  }

  if (compatinfo->ident < here->info->ident)
    here->left  = TclObj_AddComptree (interp, here->left,
				      compatinfo, result);
  else if (compatinfo->ident > here->info->ident)
    here->right = TclObj_AddComptree (interp, here->right,
				      compatinfo, result);

  return here;
}

int
TclObj_BuildComptree (Tcl_Interp *interp, TclObj_Info *me)
{
  struct _basebfsqueue {
    TclObj_Info *data;
    struct _basebfsqueue *next;
  } *bfs=NULL, *tail=NULL, *node;
  TclObj_Info *theinfo;
  int errcode=TCL_OK;
  const char **iter;


  if (me->compats)
    return TCL_OK;

  /*
   * perform BFS through the class hierarchy. start by adding
   * myself.
   * while we're add it, also add the signatures to the sigtree.
   */

  if ((bfs = tail = node = new struct _basebfsqueue) == NULL) {
    Tcl_AppendResult (interp, "out of memory in TclObj_BuildComptree\n",
		      NULL);
    errcode = TCL_ERROR;
  }
  else {
    node->data = me;
    node->next = NULL;
  }

  /*
   * perform bfs
   */

  while (bfs && !errcode) {
    me->compats = TclObj_AddComptree  (interp, me->compats, bfs->data,
				       &errcode);
    me->members = TclObj_BuildSigtree (interp, me->members, bfs->data,
				       &errcode, (bfs->data==me)?0:1);
    /*
     * add base classes of bfs->data to queue
     */

    iter = &bfs->data->parent;

    if (*iter) {
      if ((theinfo = TclObj_GetClassInfo (interp, *iter)) == NULL) {
	/*
	 * not yet available, try to find it
	 */
	if ((theinfo = TclObj_Require (interp, *iter)) == NULL) {
	  errcode = TCL_ERROR;
	  break;
	}
      }

      if ((node = new struct _basebfsqueue) == NULL) {
	Tcl_AppendResult (interp, "out of memory in TclObj_BuildComptree\n",
			  NULL);
	errcode = TCL_ERROR;
	break;
      }
      
      node->data = theinfo;
      node->next = NULL;
      
      tail->next = node;
      tail = node;
    }

    /*
     * remove head of queue
     */

    node = bfs;
    bfs  = bfs->next;
    delete node;
  }

  /*
   * if errcode != TCL_OK, there might be items left in the queue
   */

  while (bfs) {
    node = bfs;
    bfs  = bfs->next;
    delete node;
  }

  return errcode;
}

int
TclObj_IsCompatible (Tcl_Interp *interp,
		     const TclObj_Info *theinfo,
		     unsigned long oident)
{
  const TclObj_Comptree *iter=theinfo->compats;

  while (iter) {
    if (oident < iter->info->ident)
      iter = iter->left;
    else if (oident > iter->info->ident)
      iter = iter->right;
    else
      break;
  }

  if (iter == NULL)
    return 0;

  return 1;
}

int
TclObj_IsCompatible (Tcl_Interp *interp,
		     const TclObj_Info *theinfo,
		     const char *thetype)
{
  const TclObj_Info *oinfo = TclObj_GetClassInfo (interp, thetype);

  if (oinfo == NULL)
    return 0;

  return TclObj_IsCompatible (interp, theinfo, oinfo->ident);
}

static int
TclObj_DumpComptree (Tcl_Interp *interp,
		     const TclObj_Comptree *compats)
{
  if (compats) {
    Tcl_AppendElement   (interp, (char*) compats->info->name);
    TclObj_DumpComptree (interp, compats->left);
    TclObj_DumpComptree (interp, compats->right);
  }
  return TCL_OK;
}

int
TclObj_DumpComptree (Tcl_Interp *interp,
		     const TclObj_Info *theinfo)
{
  return TclObj_DumpComptree (interp, theinfo->compats);
}
